from .clean import *
from .conversion import *
from .distributed import *
from .experiment import *
from .json import *
from .logger import *
from .padding import *
from .span import *
from .timer import *
from .transformers import *



import collections
import re
import string

import torch


def normalize_answer(s):
    def remove_articles(text):
        regex = re.compile(r'\b(a|an|the)\b', re.UNICODE)
        return re.sub(regex, ' ', text)

    def white_space_fix(text):
        return ' '.join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return ''.join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()
    return white_space_fix(remove_articles(remove_punc(lower(s))))


def get_tokens(s):
    if not s:
        return []
    return normalize_answer(s).split()


def compute_exact(a_gold, a_pred):
    return int(normalize_answer(a_gold) == normalize_answer(a_pred))


def compute_f1(a_gold, a_pred):
    gold_toks = get_tokens(a_gold)
    pred_toks = get_tokens(a_pred)
    common = collections.Counter(gold_toks) & collections.Counter(pred_toks)
    num_same = sum(common.values())
    if len(gold_toks) == 0 or len(pred_toks) == 0:
        return int(gold_toks == pred_toks)
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(pred_toks)
    recall = 1.0 * num_same / len(gold_toks)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1


def get_best_span(a, b):
    # a: T(batch_size, seq_len)
    # b: T(batch_size, seq_len)
    s = a.shape[1]
    ar = torch.arange(s, device=a.device)
    score = a.unsqueeze(1) + b.unsqueeze(2) - (torch.gt(ar.unsqueeze(0) - ar.unsqueeze(1), 0).long() + torch.lt(ar.unsqueeze(0) - ar.unsqueeze(1), -15).long()).unsqueeze(0) * 1e7
    m = score.view(a.shape[0], -1).argmax(1)
    indices = torch.cat(((m % s).view(-1, 1), (m / s).view(-1, 1)), dim=1)
    confidence = torch.exp(torch.max(torch.max(score, dim=1)[0], dim=1)[0] - torch.logsumexp(torch.logsumexp(score, dim=1), dim=1))
    return indices, confidence


def list_find(a, b):
    first_matches = list()
    offset = -1
    element = b[0]
    while True:
        try:
            offset = a.index(element, offset + 1)
        except ValueError:
            break
        first_matches.append(offset)
    ret = list()
    for m in first_matches:
        if m + len(b) <= len(a):
            match = True
            for i in range(1, len(b)):
                if a[m + i] != b[i]:
                    match = False
                    break
            if match:
                ret.append([m, m + len(b) - 1])
    return ret



import json

import torch


def distributed_broadcast(args, l):
    assert type(l) == list or type(l) == dict
    if args.local_rank < 0:
        return l
    else:
        torch.distributed.barrier()
        process_number = torch.distributed.get_world_size()
        json.dump(l, open(f'tmp/{args.local_rank}.json', 'w'))
        torch.distributed.barrier()
        objs = list()
        for i in range(process_number):
            objs.append(json.load(open(f'tmp/{i}.json')))
        if type(objs[0]) == list:
            ret = list()
            for i in range(process_number):
                ret.extend(objs[i])
        else:
            ret = dict()
            for i in range(process_number):
                for k, v in objs.items():
                    assert k not in ret
                    ret[k] = v
        torch.distributed.barrier()
        return ret



class Logger:
    def __init__(self, keys, writer, logging_steps, local_rank):
        self.d = {key: list() for key in keys}
        self.step = {key: 0 for key in keys}
        self.writer = writer
        self.logging_steps = logging_steps
        self.local_rank = local_rank

    def log(self, **kwargs):
        for key in kwargs.keys():
            self.d[key].append(kwargs[key])
            if len(self.d[key]) % self.logging_steps == 0:
                if self.local_rank <= 0:
                    self.d[key] = [k for k in self.d[key] if k is not None]
                    if len(self.d[key]) > 0:
                        self.writer.add_scalar(key, sum(self.d[key]) / len(self.d[key]), self.step[key])
                self.step[key] += 1
                self.d[key] = list()



from transformers import BertConfig, BertForSequenceClassification, BertTokenizer, XLNetConfig, XLNetForSequenceClassification, XLNetTokenizer, XLMConfig, XLMForSequenceClassification, XLMTokenizer, RobertaConfig, RobertaForSequenceClassification, RobertaTokenizer, DistilBertConfig, DistilBertForSequenceClassification, DistilBertTokenizer, AlbertConfig, AlbertForSequenceClassification, AlbertTokenizer, XLMRobertaConfig, XLMRobertaForSequenceClassification, XLMRobertaTokenizer, FlaubertConfig, FlaubertForSequenceClassification, FlaubertTokenizer
transformer_classes = {
    "bert": (BertConfig, BertForSequenceClassification, BertTokenizer),
    "xlnet": (XLNetConfig, XLNetForSequenceClassification, XLNetTokenizer),
    "xlm": (XLMConfig, XLMForSequenceClassification, XLMTokenizer),
    "roberta": (RobertaConfig, RobertaForSequenceClassification, RobertaTokenizer),
    "distilbert": (DistilBertConfig, DistilBertForSequenceClassification, DistilBertTokenizer),
    "albert": (AlbertConfig, AlbertForSequenceClassification, AlbertTokenizer),
    "xlmroberta": (XLMRobertaConfig, XLMRobertaForSequenceClassification, XLMRobertaTokenizer),
    "flaubert": (FlaubertConfig, FlaubertForSequenceClassification, FlaubertTokenizer),
}


import json
import os
import random
import shutil
import sys

import IPython
import numpy as np
import torch


def set_seed(args):
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    torch.cuda.manual_seed_all(args.seed)


def experiment_init(args):
    set_seed(args)
    if args.debug:
        sys.excepthook = IPython.core.ultratb.FormattedTB(mode='Verbose', color_scheme='Linux', call_pdb=1)
    if args.local_rank == -1:
        args.device = torch.device('cuda')
    else:
        torch.cuda.set_device(args.local_rank)
        args.device = torch.device("cuda", args.local_rank)
        torch.distributed.init_process_group(backend='nccl')
    if args.local_rank > 0:
        torch.distributed.barrier()
    if args.local_rank <= 0:
        if not os.path.exists('r'):
            os.mkdir('r')
        runs = os.listdir('r')
        i = max([int(c) for c in runs], default=-1) + 1
        os.mkdir(os.path.join('r', str(i)))
        src_names = [source for source in os.listdir() if source.endswith('.py')]
        for src_name in src_names:
            shutil.copy(src_name, os.path.join('r', str(i), src_name))
        os.chdir(os.path.join('r', str(i)))
        os.mkdir('output')
        os.mkdir('tmp')
    else:
        runs = os.listdir('r')
        i = max([int(c) for c in runs])
        os.chdir(os.path.join('r', str(i)))
    if args.local_rank == 0:
        torch.distributed.barrier()
    if not args.train:
        args.epoch = 1
    if args.local_rank <= 0:
        print(args)
    print(f'Process rank: {args.local_rank}, device: {args.device}, distributed training: {bool(args.local_rank != -1)}')
    json.dump(sys.argv, open('output/args.json', 'w'))


def experiment_finish():
    json.dump(True, open('output/f.json', 'w'))
